import six
import json

if six.PY3:
    from io import StringIO
    import urllib.request as urllib
    from urllib.request import HTTPError
    from unittest import mock, TestCase
else:
    import urllib2 as urllib
    import mock
    from urllib2 import HTTPError
    from unittest import TestCase
    from cStringIO import StringIO

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from request_signer import constants
from request_signer.client.generic import Client, Response, Request, django_backend


class TestableClient(Client):

    def __init__(self):
        api_credentials = django_backend.DjangoSettingsApiCredentialsBackend(self)
        super(TestableClient, self).__init__(api_credentials)


class ClientTests(TestCase):

    def setUp(self):
        self.settings_name = 'SAMPLE_AUTH_DOMAIN'
        self.client_id_settings_name = 'SAMPLE_AUTH_CLIENT_ID'
        self.private_key_settings_name = 'SAMPLE_AUTH_PRIVATE_KEY'
        self.client = TestableClient()
        self.domain = 'http://www.sample.com'
        self.path = '/some/path/'
        self.url = self.domain + self.path
        self.client_id = 'UJSDAYFH$%U&GCKDJBSG'
        self.private_key = 'Q0TBXqjE60nfoHaQ1CZxWxt4JBUyhBPJkH1y_274tDI='
        self.client_id_param_name = 'X-TEST-CLIENT-HEADER'
        self.signature_param_name = 'X-TEST-SIGNATURE'
        self.setup_client()

        self.endpoint = '/some/endpoint'
        self.urlopen_patch = mock.patch.object(urllib, 'urlopen')
        self.urlopen = self.urlopen_patch.start()
        self.request = Request('GET', self.url, dict())

    def setup_client(self):
        setattr(self.client, 'domain_settings_name', self.settings_name)
        setattr(self.client, 'client_id_settings_name', self.client_id_settings_name)
        setattr(self.client, 'private_key_settings_name', self.private_key_settings_name)
        setattr(settings, self.settings_name, self.domain)
        setattr(settings, self.client_id_settings_name, self.client_id)
        setattr(settings, self.private_key_settings_name, self.private_key)

    def tearDown(self):
        self.urlopen_patch.stop()
        self.teardown_client()

    def teardown_client(self):
        source = settings
        self.safe_delattr(source, self.settings_name)
        self.safe_delattr(source, self.client_id_settings_name)
        self.safe_delattr(source, self.private_key_settings_name)

    def safe_delattr(self, source, name):
        try:
            delattr(source, name)
        except AttributeError:
            pass

    def get_response(self, http_method="GET", endpoint="/some/endpoint", *args, **kwargs):
        return self.client._get_response(http_method, endpoint, *args, **kwargs)

    def assert_client_raises_improperly_configured_with_attribute_name(self, client_attribute, client_property):
        delattr(self.client, client_attribute)
        with self.assertRaises(ImproperlyConfigured) as error:
            getattr(self.client, client_property)
        self.assertIn(client_attribute, str(error.exception))

    def assert_client_raises_improperty_configured_with_missing_settings_name(self, settings_name, client_property):
        delattr(settings, settings_name)
        with self.assertRaises(ImproperlyConfigured) as error:
            getattr(self.client, client_property)
        self.assertIn(settings_name, str(error.exception))

    @mock.patch('generic_request_signer.request.Request')
    def test_get_response_creates_request(self, request):
        method = 'GET'
        data = dict(right='here', some='data', this="is")
        request_kwargs = {"headers": {"Accept": "application/json"}}
        with mock.patch.object(Client, '_get_service_url') as get_url:
            get_url.return_value = 'http://my_url'
            self.get_response(method, self.endpoint, data, **request_kwargs)

        request.assert_called_once_with(
            method, 'http://my_url?{0}={1}&right=here&some=data&this=is&{2}={3}'.format(
                constants.CLIENT_ID_PARAM_NAME, self.client._client_id, constants.SIGNATURE_PARAM_NAME,
                '5oRsP9rQ-qozJlwlrV2DrC3JJitT4J8zXsuucnN6Ogg='
            ), None, **request_kwargs
        )

    @mock.patch('generic_request_signer.request.Request')
    def test_get_response_creates_request_with_json_payload(self, request):
        method = 'POST'
        data = dict(this="is", some='data', right='here')
        request_kwargs = dict(headers={"Content-Type": "application/json"})
        with mock.patch.object(Client, '_get_service_url') as get_url:
            get_url.return_value = 'http://my_url'
            self.get_response(method, self.endpoint, json.dumps(data, sort_keys=True), **request_kwargs)

        request.assert_called_once_with(
            method, 'http://my_url?{0}={1}&{2}={3}'.format(
                constants.CLIENT_ID_PARAM_NAME, self.client._client_id, constants.SIGNATURE_PARAM_NAME,
                'GQuoMFCNPBPoG736rfILRebBvlcnaj72LJU4cVSxqQo='
            ), json.dumps(data, sort_keys=True).encode('utf8'), **request_kwargs
        )

    def test_get_response_gets_url_with_endpoint(self):
        with mock.patch.object(Client, '_get_service_url') as get_url:
            get_url.return_value = 'http://url'
            self.get_response(endpoint=self.endpoint)
        get_url.assert_called_once_with(self.endpoint)

    def test_get_url_returns_url_from_settings_with_endpoint(self):
        url = self.client._get_service_url(endpoint=self.endpoint)
        self.assertEqual(self.domain + self.endpoint, url)

    def test_raises_improperly_configured_when_no_domain_settings_in_settings(self):
        settings_name = self.settings_name
        client_property = '_base_url'
        self.assert_client_raises_improperty_configured_with_missing_settings_name(settings_name, client_property)

    def test_raises_improperly_configured_when_no_url_in_settings(self):
        settings_name = self.client_id_settings_name
        client_property = '_client_id'
        self.assert_client_raises_improperty_configured_with_missing_settings_name(settings_name, client_property)

    def test_raises_improperly_configured_when_no_private_key_in_settings(self):
        settings_name = self.private_key_settings_name
        client_property = '_private_key'
        self.assert_client_raises_improperty_configured_with_missing_settings_name(settings_name, client_property)

    def test_raises_improperly_configured_when_no_domain_settings_name(self):
        client_attribute = 'domain_settings_name'
        client_property = '_base_url'
        self.assert_client_raises_improperly_configured_with_attribute_name(client_attribute, client_property)

    def test_raises_improperly_configured_when_no_client_id_settings_name(self):
        client_attribute = 'client_id_settings_name'
        client_property = '_client_id'
        self.assert_client_raises_improperly_configured_with_attribute_name(client_attribute, client_property)

    def test_raises_improperly_configured_when_no_private_key_settings_name(self):
        client_attribute = 'private_key_settings_name'
        client_property = '_private_key'
        self.assert_client_raises_improperly_configured_with_attribute_name(client_attribute, client_property)

    def test_client_id_property_returns_client_id_from_settings(self):
        self.assertEqual(self.client_id, self.client._client_id)

    def test_private_key_property_returns_private_key_from_settings(self):
        self.assertEqual(self.private_key, self.client._private_key)

    @mock.patch('generic_request_signer.request.Request')
    def test_get_raw_response_invokes_urlopen_with_request(self, request):
        self.get_response()
        self.urlopen.assert_called_once_with(request.return_value, timeout=15)

    def test_returns_raw_response_wrapped_with_response_object_when_urllib_works_as_expected(self):
        raw_response = self.get_response()
        self.assertIsInstance(raw_response, Response)
        self.assertEqual(raw_response.raw_response, self.urlopen.return_value)

    def test_raw_error_returned_wrapped_with_response_object_when_exception_thrown(self):
        expected = HTTPError(None, 500, None, None, StringIO())
        self.urlopen.side_effect = expected
        error_as_response = self.get_response()
        self.assertIsInstance(error_as_response, Response)
        self.assertEqual(expected, error_as_response.raw_response)



import six
if six.PY3:
    from unittest import mock
    from io import StringIO
    from http.client import responses
else:
    import mock
    from cStringIO import StringIO
    from httplib import responses


import json
from django import test
from request_signer.client.generic import Response


class ResponseTests(test.TestCase):

    def setUp(self):
        self.raw_response = mock.Mock()
        self.response = Response(self.raw_response)

    def test_response_requires_url_to_init(self):
        self.assertEqual(self.response.raw_response, self.raw_response)

    @mock.patch.object(Response, '_evaluate_response_code_for_success')
    def test_response_is_successful_returns_value_from_evaluate(self, evaluate_response):
        self.assertEqual(self.response.is_successful, evaluate_response.return_value)

    @mock.patch.object(Response, 'status_code', mock.Mock())
    @mock.patch.object(Response, '_evaluate_response_code_for_success')
    def test_response_is_successful_calls_evaluate_with_status_code(self, evaluate_response):
        getattr(self.response, 'is_successful')
        evaluate_response.assert_called_once_with(self.response.status_code)

    def test_bad_http_status_return_false_from_evaluate_response_code_for_success(self):
        def include_status(status):
            return status < 200 or status > 299

        self.evaluate_response_code_for_success(False, include_status)

    def test_good_http_status_return_true_from_evaluate_response_code_for_success(self):
        def include_status(status):
            return 199 < status < 300

        self.evaluate_response_code_for_success(True, include_status)

    def evaluate_response_code_for_success(self, expected, include_status):
        statuses = (status for status in responses.keys() if include_status(status))
        for response_code in statuses:
            value = self.response._evaluate_response_code_for_success(response_code)
            message = "it seems '%s' returned '%s' for some odd reason" % (response_code, value)
            self.assertEqual(expected, value, message)

    def test_status_code_returns_status_code_from_raw_response(self):
        self.raw_response.code = 201
        self.assertEqual(201, self.response.status_code)

    def test_returns_dict_of_json_data_from_response(self):
        self.raw_response.read.return_value = '{"first":"item"}'
        self.assertEqual(dict(first='item'), self.response.json)

    def test_can_read_response_multiple_times(self):
        data = '{"data": "this is the response"}'
        expected = json.loads(data)
        self.response.raw_response = StringIO(data)
        self.assertEqual(expected, self.response.json)
        self.assertEqual(expected, self.response.json)



import six

if six.PY3:
    from unittest import mock
else:
    import mock

from django import test
from request_signer.client.generic import Client
from request_signer.client.generic.django_client import DjangoClient


class DjangoClientTests(test.TestCase):

    sut_class = DjangoClient

    def setUp(self):
        self.django_backend_patch = mock.patch(
            'request_signer.client.generic.django_backend.DjangoSettingsApiCredentialsBackend'
        )
        self.django_backend = self.django_backend_patch.start()

    def tearDown(self):
        self.django_backend_patch.stop()

    def test_api_credentials_created_from_django_backend_with_the_client(self):
        with mock.patch.object(Client, '__init__'):
            sut = self.sut_class()
        self.django_backend.assert_called_once_with(sut)

    def test_invokes_super_init_with_django_backend(self):
        with mock.patch.object(Client, '__init__') as init:
            self.sut_class()
        init.assert_called_once_with(self.django_backend.return_value)



from generic_request_signer.client import Client  # noqa F401
from generic_request_signer.exceptions import WebException, HttpMethodNotAllowed  # noqa F401
from generic_request_signer.request import Request  # noqa F401
from generic_request_signer.response import Response  # noqa F401
from generic_request_signer.factory import SignedRequestFactory  # noqa F401



import json
import re

import django
import six

if six.PY3:
    from urllib.parse import unquote
    from unittest import mock
else:
    from urllib import unquote
    import mock

from django import test
from django import http
from django.test.utils import override_settings
from django.core.exceptions import ImproperlyConfigured
from apysigner import get_signature

from request_signer import constants
from request_signer.validator import SignatureValidator
from request_signer.decorators import signature_required, has_valid_signature

TEST_PRIVATE_KEY = 'abc123=='


@override_settings(API_KEYS={'apps-testclient': TEST_PRIVATE_KEY})
class SignedRequestTests(test.TestCase):

    @property
    def view(self):
        return lambda request, *args, **kwargs: http.HttpResponse("Completed Test View!")

    def get_request(self, data=None):
        return test.client.RequestFactory().get('/', data=data or {})

    def test_returns_400_response_when_request_doesnt_have_any_data(self):
        response = self.client.get('/test/')
        self.assertEqual(400, response.status_code)

    def test_returns_400_response_when_request_doesnt_have_signature_in_data(self):
        response = self.client.get('/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME))
        self.assertEqual(400, response.status_code)

    def test_adds_signature_required_attribute_to_view(self):
        self.assertFalse(hasattr(self.view, 'signature_required'))
        signed_view = signature_required(self.view)
        self.assertTrue(getattr(signed_view, 'signature_required', False))

    def test_returns_400_response_when_request_doesnt_have_client_id_in_data(self):
        response = self.client.get('/test/?{}=anythingherethatiswrong'.format(constants.SIGNATURE_PARAM_NAME))
        self.assertEqual(400, response.status_code)

    def test_returns_400_when_signature_doesnt_match(self):
        response = self.client.get(
            '/test/?{}=apps-testclient&{}=anythingherethatiswrong'.format(
                constants.CLIENT_ID_PARAM_NAME, constants.SIGNATURE_PARAM_NAME
            )
        )
        self.assertEqual(400, response.status_code)

    def test_returns_200_view_return_value_when_signature_matches(self):
        url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    def test_signature_works_when_url_contains_spaces_and_querystring_contains_escaped_chars(self):
        url = '/test/a b c/?username=test%2C&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    def test_signature_works_when_url_contains_spaces(self):
        url = '/test/a b c/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    def test_signature_works_when_url_contains_escapsed_spaces(self):
        url = '/test/a%20b%20c/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    @mock.patch('request_signer.signals.successful_signed_request.send')
    @mock.patch('apysigner.get_signature')
    def test_fires_successful_signed_request_signal_from_signature_required_decorator(self, get_signature, send_signal):
        get_signature.return_value = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        request = self.get_request(
            data={
                constants.SIGNATURE_PARAM_NAME: '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI=',
                constants.CLIENT_ID_PARAM_NAME: 'apps-testclient',
            }
        )
        instance = SignatureValidator(request)
        with mock.patch('request_signer.decorators.get_validator', mock.Mock(return_value=instance)):
            signed_view = signature_required(self.view)
            response = signed_view(request)
        self.assertEqual(200, response.status_code)
        send_signal.assert_called_once_with(sender=instance, request=request)

    @mock.patch('request_signer.signals.successful_signed_request.send')
    @mock.patch('apysigner.get_signature')
    def test_fires_successful_signed_request_signal_from_has_valid_signature(self, get_signature, send_signal):
        get_signature.return_value = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        request = self.get_request(
            data={
                constants.SIGNATURE_PARAM_NAME: '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI=',
                constants.CLIENT_ID_PARAM_NAME: 'apps-testclient',
            }
        )
        instance = SignatureValidator(request)
        with mock.patch('request_signer.decorators.get_validator', mock.Mock(return_value=instance)):
            self.assertTrue(has_valid_signature(request))
        send_signal.assert_called_once_with(sender=instance, request=request)

    @mock.patch('request_signer.signals.successful_signed_request.send')
    def test_does_not_fire_successful_signal_from_signature_required_when_invalid(self, send_signal):
        response = self.client.get(
            '/test/?{}=apps-testclient&{}=anythingherethatiswrong'.format(
                constants.CLIENT_ID_PARAM_NAME, constants.SIGNATURE_PARAM_NAME
            )
        )
        self.assertEqual(400, response.status_code)
        self.assertFalse(send_signal.called)

    @mock.patch('request_signer.signals.successful_signed_request.send')
    @mock.patch('apysigner.get_signature')
    def test_does_not_fire_successful_signal_from_has_valid_signature_when_invalid(self, get_signature, send_signal):
        get_signature.return_value = 'ABCDEFGHIJKLMNOPQRSTUVWXYZFtYkCdi4XAc-vOLtI='
        request = self.get_request(
            data={
                constants.SIGNATURE_PARAM_NAME: '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI=',
                constants.CLIENT_ID_PARAM_NAME: 'apps-testclient',
            }
        )
        self.assertFalse(has_valid_signature(request))
        self.assertFalse(send_signal.called)

    def test_calls_create_signature_properly_with_get_data(self):
        url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    def test_calls_create_signature_properly_with_get_data_and_with_an_at_symbol(self):
        url = '/test/?username=test@example.com&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    @mock.patch('apysigner.get_signature')
    def test_calls_create_signature_properly_with_no_content_type(self, get_signature):
        signature = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        get_signature.return_value = signature
        request = test.client.RequestFactory().get(
            "/my/path/?{}={}&{}={}".format(
                constants.CLIENT_ID_PARAM_NAME,
                'apps-testclient',
                constants.SIGNATURE_PARAM_NAME,
                signature,
            )
        )
        if 'CONTENT_TYPE' in request.META:
            del request.META['CONTENT_TYPE']
        signed_view = signature_required(self.view)
        signed_view(request)
        call_url = unquote(request.get_full_path())
        call_url = re.sub(r'&__signature={}$'.format(signature), '', call_url, count=1)
        get_signature.assert_called_once_with(TEST_PRIVATE_KEY, call_url, {})

    @mock.patch('apysigner.get_signature')
    def test_json_is_properly_parsed_into_signature(self, get_signature):
        signature = 'QEw8WN5YzbWlct5ZXH3GIumeiL8m4NErPtXOz_jWexc='
        url = "/my/path/?{}={}&{}={}".format(
            constants.CLIENT_ID_PARAM_NAME,
            'apps-testclient',
            constants.SIGNATURE_PARAM_NAME,
            signature,
        )
        data = {'our': 'data', 'goes': 'here'}
        json_string = json.dumps(data)
        request = test.client.RequestFactory().post(url, data=json_string, content_type="application/json")
        signed_view = signature_required(self.view)
        signed_view(request)
        get_signature.assert_called_with(TEST_PRIVATE_KEY, '/my/path/?__client_id=apps-testclient', json_string)

    def test_json_payload_is_valid(self):
        json_string = json.dumps({'our': 'data', 'goes': 'here'})
        url = '/test/?username=test@example.com&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, json_string)
        response = self.client.post(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            json_string,
            content_type="application/json"
        )
        self.assertEqual(200, response.status_code)

    @mock.patch('apysigner.get_signature')
    def test_calls_create_signature_properly_with_post_data(self, get_signature):
        signature = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        get_signature.return_value = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        url = "/my/path/?{}={}&{}={}".format(
            constants.CLIENT_ID_PARAM_NAME,
            'apps-testclient',
            constants.SIGNATURE_PARAM_NAME,
            signature,
        )
        request = test.client.RequestFactory().post(url, data={'username': 'tester'})
        signed_view = signature_required(self.view)
        signed_view(request)
        get_signature.assert_called_once_with(
            TEST_PRIVATE_KEY, '/my/path/?__client_id=apps-testclient', dict(request.POST)
        )

    def test_post_data_payload_is_valid(self):
        data = {'username': ['tester']}
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.post('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature), data=data)
        self.assertEqual(200, response.status_code)

    @mock.patch('apysigner.get_signature')
    def test_does_not_create_signature_with_multivalue_dict_to_prevent_data_loss(self, get_signature):
        signature = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        get_signature.return_value = '4ZAQJqmWE_C9ozPkpJ3Owh0Z_DFtYkCdi4XAc-vOLtI='
        url = "/my/path/?{}={}&{}={}".format(
            constants.CLIENT_ID_PARAM_NAME,
            'apps-testclient',
            constants.SIGNATURE_PARAM_NAME,
            signature,
        )
        request = test.client.RequestFactory().post(url, data={'usernames': ['t1', 't2', 't3']})
        signed_view = signature_required(self.view)
        signed_view(request)

        expected_url = re.sub(r'&__signature={}$'.format(signature), '', unquote(request.get_full_path()), count=1)
        get_signature.assert_called_once_with(TEST_PRIVATE_KEY, expected_url, dict(request.POST))
        posted_data = get_signature.mock_calls[0][1][2]
        self.assertEqual([('usernames', ['t1', 't2', 't3'])], list(posted_data.items()))

    def test_post_multivalue_data_payload_is_valid(self):
        data = {'username': ['tester', 'billyjean']}
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.post('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature), data=data)
        self.assertEqual(200, response.status_code)

    def test_signed_views_are_csrf_exempt(self):
        signed_view = signature_required(self.view)
        self.assertTrue(getattr(signed_view, 'csrf_exempt', False))

    @override_settings(ALLOW_UNSIGNED_REQUESTS=True)
    def test_returns_200_when_signature_doesnt_match_but_allow_unsigned_is_true(self):
        response = self.client.get(
            '/test/?{}=apps-testclient&{}=anythingherethatiswrong'.format(
                constants.CLIENT_ID_PARAM_NAME, constants.SIGNATURE_PARAM_NAME
            )
        )
        self.assertEqual(200, response.status_code)

    def test_put_requests_still_use_request_body(self):
        url = '/asdf/'
        request = test.client.RequestFactory().post(
            url, data={'usernames': ['t1', 't2', 't3']}, **{
                'REQUEST_METHOD': "PUT"
            }
        )
        request_without_data = test.client.RequestFactory().post(url, data={}, **{'REQUEST_METHOD': "PUT"})
        validator1 = SignatureValidator(request)
        validator2 = SignatureValidator(request_without_data)
        self.assertNotEqual(validator1.request_data, validator2.request_data)

    def test_put_json_is_valid(self):
        data = json.dumps({'username': ['tester', 'billyjean']})
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.put(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            data=data,
            content_type="application/json"
        )
        self.assertEqual(200, response.status_code)

    def test_patch_requests_still_use_request_body(self):
        url = '/asdf/'
        request = test.client.RequestFactory().post(
            url, data={'usernames': ['t1', 't2', 't3']}, **{
                'REQUEST_METHOD': "PATCH"
            }
        )
        request_without_data = test.client.RequestFactory().post(url, data={}, **{'REQUEST_METHOD': "PATCH"})
        validator1 = SignatureValidator(request)
        validator2 = SignatureValidator(request_without_data)
        self.assertNotEqual(validator1.request_data, validator2.request_data)

    def test_patch_json_is_valid(self):
        data = json.dumps({'username': ['tester', 'billyjean']})
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.patch(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            data=data,
            content_type="application/json"
        )
        self.assertEqual(200, response.status_code)

    def test_patch_requests_return_dict_of_multipart_form_data(self):
        request = test.client.RequestFactory().post(
            '/asdf/', data={'usernames': ['t1', 't2', 't3']}, **{
                'REQUEST_METHOD': "PATCH"
            }
        )

        if django.get_version() >= "3.2":
            expected = {
                '--BoUnDaRyStRiNg\r\nContent-Disposition: form-data; name': [
                    '"usernames"\r\n\r\nt1\r\n--BoUnDaRyStRiNg\r\nContent-Disposition: form-data; '
                    'name="usernames"\r\n\r\nt2\r\n--BoUnDaRyStRiNg\r\nContent-Disposition: form-data; '
                    'name="usernames"\r\n\r\nt3\r\n--BoUnDaRyStRiNg--\r\n'
                ]}
        else:
            expected = {
                u'--BoUnDaRyStRiNg\r\nContent-Disposition: form-data': [u''],
                u' name': [
                    u'"usernames"\r\n\r\nt1\r\n--BoUnDaRyStRiNg\r\nContent-Disposition: form-data',
                    u'"usernames"\r\n\r\nt2\r\n--BoUnDaRyStRiNg\r\nContent-Disposition: form-data',
                    u'"usernames"\r\n\r\nt3\r\n--BoUnDaRyStRiNg--\r\n'
                ]
            }

        self.assertEqual(expected, SignatureValidator(request).request_data)

    @mock.patch('apysigner.get_signature')
    def test_json_api_is_properly_parsed_into_signature(self, get_signature):
        signature = 'QEw8WN5YzbWlct5ZXH3GIumeiL8m4NErPtXOz_jWexc='
        url = "/my/path/?{}={}&{}={}".format(
            constants.CLIENT_ID_PARAM_NAME,
            'apps-testclient',
            constants.SIGNATURE_PARAM_NAME,
            signature,
        )
        data = {'our': 'data', 'goes': 'here'}
        json_string = json.dumps(data)
        request = test.client.RequestFactory().post(url, data=json_string, content_type="application/vnd.api+json")
        signed_view = signature_required(self.view)
        signed_view(request)
        get_signature.assert_called_with(TEST_PRIVATE_KEY, '/my/path/?__client_id=apps-testclient', json_string)

    def test_json_api_payload_is_valid(self):
        json_string = json.dumps({'our': 'data', 'goes': 'here'})
        url = '/test/?username=test@example.com&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, json_string)
        response = self.client.post(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            json_string,
            content_type="application/vnd.api+json"
        )
        self.assertEqual(200, response.status_code)

    def test_put_json_api_is_valid(self):
        data = json.dumps({'username': ['tester', 'billyjean']})
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.put(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            data=data,
            content_type="application/vnd.api+json"
        )
        self.assertEqual(200, response.status_code)

    def test_patch_json_api_is_valid(self):
        data = json.dumps({'username': ['tester', 'billyjean']})
        url = '/test/?{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url, payload=data)
        response = self.client.patch(
            '{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature),
            data=data,
            content_type="application/vnd.api+json"
        )
        self.assertEqual(200, response.status_code)

    def test_does_not_call_database_per_request(self):
        url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature(TEST_PRIVATE_KEY, url)
        with self.assertNumQueries(0):
            response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
            self.assertEqual(200, response.status_code)

    def test_returns_200_when_using_api_keys(self):
        url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature('abc123==', url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(200, response.status_code)

    @override_settings(API_KEYS={'apps-no': TEST_PRIVATE_KEY})
    def test_returns_400_when_using_api_keys_that_do_not_match(self):
        url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature('abc123==', url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(400, response.status_code)
        url = '/test/?username=test&{}=app-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
        signature = get_signature('123==', url)
        response = self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))
        self.assertEqual(400, response.status_code)


class NoSettingsClass(test.TestCase):

    def test_request_fails_without_correct_settings(self):
        with self.assertRaises(ImproperlyConfigured):
            url = '/test/?username=test&{}=apps-testclient'.format(constants.CLIENT_ID_PARAM_NAME)
            signature = get_signature('abc123==', url)
            self.client.get('{}&{}={}'.format(url, constants.SIGNATURE_PARAM_NAME, signature))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='AuthorizedClient',
            fields=[
                ('client_id', models.CharField(max_length=20, serialize=False, primary_key=True)),
                ('private_key', models.CharField(default='', max_length=100)),
                ('is_active', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
            options={},
            bases=(models.Model, ),
        ),
    ]




import six

if six.PY3:
    from unittest import mock
else:
    import mock

from django import test
from request_signer.client.generic import Response, WebException

from django.test.utils import override_settings
from request_signer.client.generic.rest import BaseDjangoRestClient


class BaseDjangoRestClientTests(test.TestCase):

    def setUp(self):
        self.sut = BaseDjangoRestClient()
        self.sut.BASE_API_ENDPOINT = "/api/"

    def get_mock_response(self, **kwargs):
        return mock.MagicMock(Response, autospec=True, **kwargs)

    def test_build_endpoint_returns_endpoint_when_only_group_provided(self):
        endpoint = self.sut.build_endpoint("1234")
        self.assertEqual("/api/1234/", endpoint)

    def test_build_endpoint_returns_endpoint_when_group_and_item_provided(self):
        endpoint = self.sut.build_endpoint("1234", "item-detail")
        self.assertEqual("/api/1234/item-detail/", endpoint)

    def test_get_json_response_gets_response_with_accept_header(self):
        endpoint = self.sut.build_endpoint("1234", "item-detail")
        data = {"some_data": "lives here"}

        with mock.patch.object(self.sut, "_get_response") as get_response:
            self.sut._get_json_response("GET", endpoint, data)
        get_response.assert_called_once_with("GET", endpoint, data, headers={"Accept": "application/json"})

    def test_get_list_issues_get_json_response_for_endpoint(self):
        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            self.sut.get_list("1234")
        get_response.assert_called_once_with("GET", "/api/1234/")

    def test_get_list_returns_json_response_when_successful(self):
        response = self.get_mock_response()

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.get_list("1234")
        self.assertEqual(response.json, result)

    def test_get_list_returns_json_response_when_404(self):
        response = self.get_mock_response(status_code=404, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.get_list("1234")
        self.assertEqual(response.json, result)

    def test_get_list_raises_web_exception_when_not_successful(self):
        response = self.get_mock_response(status_code=500, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            with self.assertRaises(WebException) as e:
                self.sut.get_list("1234")
        self.assertEqual(str(response.read.return_value), str(e.exception.message))

    def test_get_item_issues_get_json_response_for_endpoint(self):
        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            self.sut.get_item("1234", "pk-3")
        get_response.assert_called_once_with("GET", "/api/1234/pk-3/")

    def test_get_item_returns_json_response_when_successful(self):
        response = self.get_mock_response()

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.get_item("1234", "pk-3")
        self.assertEqual(response.json, result)

    def test_get_item_returns_json_response_when_404(self):
        response = self.get_mock_response(status_code=404, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.get_item("1234", "pk-3")
        self.assertEqual(response.json, result)

    def test_get_item_raises_web_exception_when_not_successful(self):
        response = self.get_mock_response(status_code=500, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            with self.assertRaises(WebException) as e:
                self.sut.get_item("1234", "pk-3")
        self.assertEqual(str(response.read.return_value), str(e.exception.message))

    def test_create_issues_get_json_response_for_endpoint(self):
        data = {'some_data': 'to send'}
        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            self.sut.create("1234", **data)
        get_response.assert_called_once_with("POST", "/api/1234/", data=data)

    def test_create_returns_json_response_when_successful(self):
        response = self.get_mock_response()
        data = {'some_data': 'to send'}

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.create("1234", **data)
        self.assertEqual(response.json, result)

    def test_create_raises_web_exception_when_not_successful(self):
        response = self.get_mock_response(status_code=400, is_successful=False)
        data = {'some_data': 'to send'}

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            with self.assertRaises(WebException) as e:
                self.sut.create("1234", **data)
        self.assertEqual(str(response.read.return_value), str(e.exception.message))

    def test_update_issues_get_json_response_for_endpoint(self):
        data = {'some_data': 'to send'}
        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            self.sut.update("1234", "pk-3", **data)

        expected_data = dict(data, _method="PUT")
        get_response.assert_called_once_with("POST", "/api/1234/pk-3/", data=expected_data)

    def test_update_returns_json_response_when_successful(self):
        response = self.get_mock_response()
        data = {'some_data': 'to send'}

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.update("1234", "pk-3", **data)
        self.assertEqual(response.json, result)

    def test_update_raises_web_exception_when_not_successful(self):
        response = self.get_mock_response(status_code=404, is_successful=False)
        data = {'some_data': 'to send'}

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            with self.assertRaises(WebException) as e:
                self.sut.update("1234", "pk-3", **data)
        self.assertEqual(str(response.read.return_value), str(e.exception.message))

    def test_delete_issues_get_json_response_for_endpoint(self):
        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            self.sut.delete("1234", "pk-3")
        get_response.assert_called_once_with("POST", "/api/1234/pk-3/", data={"_method": "DELETE"})

    def test_delete_returns_json_response_when_successful(self):
        response = self.get_mock_response()

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.delete("1234", "pk-3")
        self.assertEqual(response.json, result)

    def test_delete_returns_json_response_when_404(self):
        response = self.get_mock_response(status_code=404, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            result = self.sut.delete("1234", "pk-3")
        self.assertEqual(response.json, result)

    def test_delete_raises_web_exception_when_not_successful(self):
        response = self.get_mock_response(status_code=500, is_successful=False)

        with mock.patch.object(self.sut, "_get_json_response") as get_response:
            get_response.return_value = response
            with self.assertRaises(WebException) as e:
                self.sut.delete("1234", "pk-3")
        self.assertEqual(str(response.read.return_value), str(e.exception.message))


class BaseDjangoRestClientInitTests(test.TestCase):

    @override_settings(TEST_DOMAIN='my_domain')
    @override_settings(TEST_CLIENT_ID='my_client_id')
    @override_settings(TEST_PRIVATE_KEY='my_private_key')
    def test_uses_django_settings_by_default_for_api_credentials(self):

        class SomeClient(BaseDjangoRestClient):
            domain_settings_name = 'TEST_DOMAIN'
            client_id_settings_name = 'TEST_CLIENT_ID'
            private_key_settings_name = 'TEST_PRIVATE_KEY'

        rest_client = SomeClient()

        self.assertEqual('my_domain', rest_client._base_url)
        self.assertEqual('my_client_id', rest_client._client_id)
        self.assertEqual('my_private_key', rest_client._private_key)

    def test_will_use_provided_settings_when_available(self):

        class SomeProvider(object):
            base_url = "my_domain"
            client_id = "my_client_id"
            private_key = "my_private_key"

        class SomeClient(BaseDjangoRestClient):
            pass

        provider = SomeProvider()
        rest_client = SomeClient(provider)

        self.assertEqual('my_domain', rest_client._base_url)
        self.assertEqual('my_client_id', rest_client._client_id)
        self.assertEqual('my_private_key', rest_client._private_key)



from django import test
from request_signer.client.generic import Request, HttpMethodNotAllowed


class RequestTests(test.TestCase):

    def get_request(self, http_method, url="http://some.domain.com"):
        return Request(http_method, url, None)

    def assert_sets_http_method(self, http_method):
        request = self.get_request(http_method)
        self.assertEqual(http_method, request.get_method())

    def test_sets_method_to_get_when_passed_to_init(self):
        self.assert_sets_http_method('GET')

    def test_sets_method_to_post_when_passed_to_init(self):
        self.assert_sets_http_method('POST')

    def test_sets_method_to_put_when_passed_to_init(self):
        self.assert_sets_http_method('PUT')

    def test_sets_method_to_delete_when_passed_to_init(self):
        self.assert_sets_http_method('DELETE')

    def test_sets_method_to_head_when_passed_to_init(self):
        self.assert_sets_http_method('HEAD')

    def test_sets_method_to_options_when_passed_to_init(self):
        self.assert_sets_http_method('OPTIONS')

    def test_sets_method_to_trace_when_passed_to_init(self):
        self.assert_sets_http_method('TRACE')

    def test_raises_http_response_not_allowed_when_other_http_method(self):
        with self.assertRaises(HttpMethodNotAllowed):
            self.get_request('GARBAGE')



import functools

from django import http
from django.conf import settings
from django.views.decorators.csrf import csrf_exempt

from request_signer import validator


def signature_required(func):
    """
    Decorator to require a signed request.

    :param func:
        The view function that requires a signature.

    :returns:
        A new view function wrapped to ensure it is properly signed.
    """

    @csrf_exempt
    @functools.wraps(func)
    def _wrap(request, *args, **kwargs):
        """
        Returns bad request when:
          - no signature
          - no client
          - signature doesnt match
        """
        if has_valid_signature(request) or allow_unsigned_requests():
            return func(request, *args, **kwargs)
        else:
            return http.HttpResponseBadRequest()

    _wrap.signature_required = True
    return _wrap


def allow_unsigned_requests():
    return getattr(settings, 'ALLOW_UNSIGNED_REQUESTS', False)


def get_validator(request):
    return validator.SignatureValidator(request)


def has_valid_signature(request):
    return get_validator(request).has_valid_signature()



from setuptools import setup, find_packages

setup(
    name='django-request-signer',
    version='5.1.3',
    author='imtapps',
    url='https://github.com/imtapps/django-request-signer',
    description="A python library for signing http requests.",
    long_description=open('README.rst', 'r').read(),
    install_requires=open('requirements/dist.txt').read().split("\n"),
    packages=find_packages(exclude=("example", "server")),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Framework :: Django",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.2",
    ],
)



import six
from collections import namedtuple
from django.conf import settings
from django.http import QueryDict
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from generic_request_signer.check_signature import check_signature

from request_signer import constants
from request_signer.signals import successful_signed_request

if six.PY3:
    from urllib.parse import unquote
else:
    from urllib import unquote


class SignatureValidator(object):

    def __init__(self, request):
        self.request = request

    def has_valid_signature(self):
        self._fire_signal_when_signature_valid()
        return self.signature_was_valid

    def _fire_signal_when_signature_valid(self):
        if self.signature_was_valid:
            successful_signed_request.send(sender=self, request=self.request)

    @cached_property
    def signature_was_valid(self):
        if self.client:
            result = check_signature(self.signature, self.client.private_key, self.url_path, self.request_data)
            return result or check_signature(
                self.signature, self.client.private_key, unquote(self.url_path), self.request_data
            ) or self.unquote_base_url

    @property
    def unquote_base_url(self):
        url, query = self.url_path.split('?')
        return check_signature(
            self.signature, self.client.private_key, '{}?{}'.format(unquote(url), query), self.request_data
        )

    @property
    def signature(self):
        return self.request.GET.get(constants.SIGNATURE_PARAM_NAME)

    @property
    def client_id(self):
        return self.request.GET.get(constants.CLIENT_ID_PARAM_NAME)

    @property
    def url_path(self):
        return self.request.get_full_path()

    @cached_property
    def client(self):
        if not self.signature or not self.client_id:
            return False
        if getattr(settings, 'API_KEYS', None):
            Client = namedtuple('client', ['private_key'])
            return Client(settings.API_KEYS.get(self.client_id, ''))
        raise ImproperlyConfigured('API_KEYS not found in settings')

    @property
    def request_data(self):
        if self.request.META.get('CONTENT_TYPE') in ['application/json', 'application/vnd.api+json']:
            request_data = self.request.body
        elif self.request.method.lower() in ['patch', 'put']:
            request_data = dict(QueryDict(self.request.body, encoding='utf-8'))
        else:
            request_data = dict(self.request.POST)
        return request_data
